/*-------------------------------------------------------------------*/
/* SNMP (Simple Network Monitoring Protocol) Receiver.    (RFC 1157) */
/* ---------------------------------------------------               */
/* Copyright  2002, Marc Niegowski                                  */
/* Systems Programmer At Large                                       */
/* Connectivity, Inc.                                                */
/* 23 W. Fourth Street                                               */
/* Media,  PA  19063-2805                                            */
/*                                                                   */
/* Phone: (610) 566-0227                                             */
/* Fax:   (610) 566-0641                                             */
/*                                                                   */
/* Email:  Marc@Tech-Center.com                                      */
/* HTTP:   http://www.marcsweb.com                                   */
/* FTP:    ftp://ftp.marcsweb.com                                    */
/* TELNET: marcsweb.com:23                                           */
/*-------------------------------------------------------------------*/

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <iostream.h>
#include <iomanip.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef   AIX
#include <sys/libcsys.h>
#endif

#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <syslog.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

#include <snmplog.h>

/*-------------------------------------------------------------------*/
/* Entry Point.                                                      */
/*-------------------------------------------------------------------*/

int main(int argc, char **argv, char **envp)
{
    char        *c;

    hostent     *host;              // UDP Receiver host entry (Read Only).
    sockaddr_in  serv;              // UDP Server socket Internet address.
 
    c = Copyright;

    ParseArgs(argc, argv);          // Parse command line arguments.
    
    if (DaemonFlag)                 // Fork off the daemon.
        StartDaemon();
    
    int    x = 0;                   // Set up the signal handlers.
    
    while (SIGNAL_HANDLERS[x])
        signal(SIGNAL_HANDLERS[x++], SigHandler);
        

    SavePID(getpid());              // Save the PID of the current process.
        
    StatusMsg(MSG_START);           // Display status message.
    
    GetHost("0.0.0.0", &host);      // Get Receiver host entry.
    OpenSocks(&iUDPSock);           // Open a UDP socket.
    BindSocks(&iUDPSock,            // Bind the socket to the host.
              &serv, host);    
    Receiver(iUDPSock);             // Receive UDP Datagrams.
    CloseSocks(iUDPSock);           // Close UDP socket.
    
    StatusMsg(MSG_STOP);            // Display status message.
    
    UnlinkPID();                    // Remove the PID file for this process.
    exit(0);                        // Return to caller with RC=0.
}

/*-------------------------------------------------------------------*/
/* Start Daemon (By terminating parent).                             */
/*                                                                   */
/* Call With:   Nothing.                                             */
/*                                                                   */
/* Returns:     void                                                 */
/*-------------------------------------------------------------------*/

void StartDaemon(void)
{
    int x = fork();
        
    if (x != 0)
    {
        if (x == -1)
        {
            int e = errno;
              
            cout << Messages[MSG_DAEMON_FAIL]
                 << "\n"
                 << Messages[MSG_ERROR_CODE]
                 << e
                 << ".\n"
                 << strerror(e)
                 << '.'
                 << endl;

            ErrorMsg(MSG_DAEMON_FAIL, e);                     
            exit(16);
        }
        else
        {
            cout << Messages[MSG_DAEMON_START]
                 << endl;
        
            openlog("SNMPLog", LOG_PID, LOG_DAEMON);
            syslog(LOG_DAEMON | LOG_NOTICE,
                "%s", Messages[MSG_DAEMON_START]);

            closelog();
            exit(0);
        }
    }
}

/*-------------------------------------------------------------------*/
/* Save the PID (Process ID of our process).                         */
/*                                                                   */
/* Call With:   int    - The PID of our process.                     */
/*                                                                   */
/* Returns:     void                                                 */
/*-------------------------------------------------------------------*/

void SavePID(int p)
{
    FILE   *f;

    int     m;
    
    m = umask(0177);
    
    f = fopen(PIDPATH, "r");
    
    if (f)
    {
        if (!DaemonFlag)
            cout << RunMessage
                 << endl;

        openlog("SNMPLog", LOG_PID, LOG_DAEMON);
        syslog(LOG_DAEMON | LOG_ERR, RunMessage + 1);
        
        fclose(f);
        closelog();
        exit(16);
    }
    
    f = fopen(PIDPATH, "w");
    
    if (!f)
    {
        int     e;
        
        e = errno;
        
        if (!DaemonFlag)
            cout << Messages[MSG_PID_CREATE_FAIL]
                 << '\n'
                 << Messages[MSG_ERROR_CODE]
                 << e
                 << ".\n"
                 << strerror(e)
                 << '.'
                 << endl;

        ErrorMsg(MSG_PID_CREATE_FAIL, e);
        exit(16);
    }
    
    fprintf(f, "%d\n", p);
    fclose(f);
}

/*-------------------------------------------------------------------*/
/* Unlink the PID (Process ID of our process) file.                  */
/*                                                                   */
/* Call With:   Nothing.                                             */
/*                                                                   */
/* Returns:     void                                                 */
/*-------------------------------------------------------------------*/

void UnlinkPID(void)
{
    int     e;
    
    if (unlink(PIDPATH) == -1)
    {
        e = errno;
    
        if (!DaemonFlag)
            cout << Messages[MSG_PID_UNLINK_FAIL]
                 << '\n'
                 << Messages[MSG_ERROR_CODE]
                 << e
                 << ".\n"
                 << strerror(e)
                 << '.'
                 << endl;

        ErrorMsg(MSG_PID_UNLINK_FAIL, e); 
    }
}

/*-------------------------------------------------------------------*/
/* Read the contents of the PID file.                                */
/*                                                                   */
/* Call With:  void                                                  */
/*                                                                   */
/* Returns:    int     - The PID or 0.                               */
/*-------------------------------------------------------------------*/

int ReadPID(void)
{
    FILE   *f;
    int     p;
    char    b[17];

    f = fopen(PIDPATH, "r");
                
    if (!f)
    {
       cout << Messages[MSG_PID_OPEN_FAIL]
            << '\n'
            << Messages[MSG_ERROR_CODE]
            << errno
            << ".\n"
            << strerror(errno)
            << '.'
            << endl;
       exit(16);
    }
        
    if (fgets(b, sizeof(b) - 1, f))
        p = atoi(b);
    else
        p = 0;

    fclose(f);
    
    return p;        
}

/*-------------------------------------------------------------------*/
/* Get UDP Receiver Host (Local Machine).                            */
/*                                                                   */
/* Call With:  char    - a pointer to an internet style IP)          */
/*                       (x.x.x.x) 4 Octets in dot notation.         */
/*             hostent - a pointer to a hostent entry for this       */
/*                       machine.                                    */
/*                                                                   */
/* Returns:    void                                                  */
/*-------------------------------------------------------------------*/

void GetHost(char *ip, hostent **host)
{
    in_addr IP;                     // Internet Address Structure.

    IP.s_addr = inet_addr(ip);      // Convert IP to unsigned long integer.

    StatusMsg(MSG_RCVR_FOUND);      // SNMP Receiver found message.
}

/*-------------------------------------------------------------------*/
/* Open UDP Receiver Socket.                                         */
/*                                                                   */
/* Call With:  int     - a pointer to a socket file descriptor.      */
/*                                                                   */
/* Returns:    void                                                  */
/*-------------------------------------------------------------------*/

void OpenSocks(int *sock)
{                                   // Allocate the UDP Receiver socket.
    *sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

    if (*sock == -1)                // Test socket.
    {
        error = errno;              // Get error number.
        SockErr(ERR_SOCKET_OPEN);   // Socket open failure message.
        exit(16);                   // Return with RC=16.
    }
    else                            // Socket successfully opened message.
        StatusMsg(MSG_SOCKET_OPENED);
}

/*-------------------------------------------------------------------*/
/* Bind UDP Receiver Socket.                                         */
/*                                                                   */
/* Call With:  int         - a pointer to a socket file descriptor.  */
/*             sockaddr_in - a pointer to assign to this socket.     */
/*             hostent     - a pointer to the host entry table.      */
/*                                                                   */
/* Returns:    void                                                  */
/*-------------------------------------------------------------------*/

void BindSocks(int *sock, sockaddr_in *serv, hostent *host)
{
                                    // Clear the server address area.
    memset(serv, 0, sizeof(sockaddr_in));
    
    serv->sin_family = AF_INET;     // Set the host & server internet address.
    serv->sin_port   = htons(162);  // Host to server comm. on port 162.

    if (bind(*sock, (struct sockaddr *) serv, sizeof(sockaddr_in))) 
    {
        error = errno;              // Get error number.
        SockErr(ERR_SOCKET_BIND);   // Bind failure message.
        exit(16);                   // Exit with RC=16.
    }
    else                            // Socket successfully bound to port 162.
        StatusMsg(MSG_SOCKET_BOUND);
}

/*-------------------------------------------------------------------*/
/* Receive UDP Datagrams on port 162.                                */
/*                                                                   */
/* Call With:  int     - a socket file descriptor.                   */
/*                                                                   */
/* Returns:    void                                                  */
/*-------------------------------------------------------------------*/

void Receiver(int sock)
{                                   // Datagram buffer.
    char        rxBuff[BUFFER_SIZE + 1];
    char        ch;                 // STDIN (Keyboard) byte.

    fd_set      fds, kb, cmd;       // Socket file descriptor set table.
    sockaddr    from;               // Sender socket address.
    socklen_t   fromlen = sizeof(from);

    time_t      tm;                 // Long integer buffer for time.
    char        ts[48];             // Buffer to hold formatted time.
    
    timeval     tv;                 // Select time-out value.
    termios     newtty; //, savetty;
    
    if (!DaemonFlag)
    {    
    
        if (tcgetattr(0, &savetty) == -1)
        {
            StatusMsg(MSG_TTY_ADDR_FAILED);
            exit(16);
        }
    
        newtty = savetty;
    
        newtty.c_lflag    &= ~ICANON;
        newtty.c_lflag    &= ~ECHO;
        newtty.c_cc[VMIN]  = 1;
        newtty.c_cc[VTIME] = 0;
    
        if (tcsetattr(0, TCSAFLUSH, &newtty) == -1)
        {
            StatusMsg(MSG_TTY_RAW_FAILED);
            exit(16);
        }
    }
    else
    {
        OpenMsgFile();
        CreatePipe();
        
        iCMDfd = open(CMDPATH, O_RDONLY | O_NONBLOCK);
        
        if (iCMDfd == -1)
        {
            ErrorMsg(MSG_PIPE_OP_FAIL, errno);
            SigHandler(0);
        }
    }
    
    FD_ZERO(&fds);                  // Clear socket/pipe file descriptor set.
    FD_ZERO(&kb);                   // Clear STDIN file descriptor set.
    FD_ZERO(&cmd);                  // CLear the command pipe descriptor set.

    while (true)                    // Begin receiving datagrams.
    {
        int x;
        
        x = 0;
        
        if (!FD_ISSET(sock, &fds))  // If the socket is not a member of the...
             FD_SET(sock, &fds);    // ...select set, add the socket to the set.
                                    // Wait for a datagram to arrive.
        tv.tv_sec  = 1;             // 1 second time-out value.
        tv.tv_usec = 0;             // 0 milliseconds.

        if ((x = select(sock + 1, &fds, NULL, NULL, &tv)))
        {                           // Check for socket error.
            if (x == -1)            // Socket error message (ignore datagram).
                SockErr(ERR_SOCKET_SELECT);
            else
            {                       // Receive datagram from port 162...
                                    // ...or test command pipe.

                if (FD_ISSET(sock, &fds))
                {
                    if ((x = recvfrom(sock, rxBuff, sizeof(rxBuff)-1,
                         0, &from,(socklen_t *) &fromlen)))
                    {               // Check for socket error.
                        if (x == -1)// Socket error message (ignore datagram).
                            SockErr(ERR_SOCKET_RECEIVE);
                        else
                        {           // Display time and raw data.
                                    // Sometimes the Linksys router,
                                    // depending on the firmware revision
                                    // doesn't terminate its message
                                    // with a newline character.

                            rxBuff[x] = '\0';

                            x = strlen(rxBuff + RAW_DATA);


                            if (rxBuff[RAW_DATA + x - 1] != '\n')
                            {
                               if (x < (BUFFER_SIZE - RAW_DATA - 1))
                                   strcat(rxBuff + RAW_DATA, "\n");
                               else
                                   rxBuff[RAW_DATA + x - 1] = '\n';
                            }

                            time(&tm);
                            
                            strftime(ts, sizeof(ts), "%x %X",
                                localtime(&tm));
                                  
                            if (!DaemonFlag)
                            {
                                cout << ts
                                     << ' '
                                     << (rxBuff[RAW_DATA] != '@' ?
                                         rxBuff + RAW_DATA       :
                                         rxBuff + RAW_DATA + 1)
                                     <<  flush;
                            }
                            else
                            fprintf(MsgFile, "%s %s", ts,
                                   (rxBuff[RAW_DATA] != '@' ?
                                    rxBuff + RAW_DATA       :
                                    rxBuff + RAW_DATA + 1));

                            iCtr++;
                        }
                    }
                }
            }
        }
        
        if (!DaemonFlag)
        {
            FD_SET(0, &kb);         // Add the STDIN file descriptor.
            tv.tv_sec  = 0;         // 0 seconds time-out value.
            tv.tv_usec = 5;         // 5 milliseconds.
            if ((x = select(1, &kb, NULL, NULL, &tv)))
                if ((x = read(0, &ch, 1)))
                    if (TTYInput(ch))
                        break;
        }
        else
        {
            FD_SET(iCMDfd, &cmd);
            tv.tv_sec = 1;
            tv.tv_usec = 0;
            x = 0;

            if ((x = select(iCMDfd + 1, &cmd, NULL, NULL, &tv)))
            {
                if (x == -1)
                    ErrorMsg(MSG_PIPE_SL_FAIL, errno);
                else
                {
                    if ((x = read(iCMDfd, &ch, 1)))
                    {
                        if (x == -1)
                            ErrorMsg(MSG_PIPE_RD_FAIL, errno);
                        else
                        {
                            if (ch == 'F' || ch == 'f')
                            {
                                if (fflush(MsgFile))
                                    ErrorMsg(MSG_LOG_FLUSH_FAIL, errno);
                                else
                                    if (SyslogFlag & DaemonFlag)
                                        StatusMsg(MSG_LOG_FLUSHED);
                             }
                         }
                    }
                }
            }
        }
    }
    
    if (!DaemonFlag)                // Reset the TTY.
        if (tcsetattr(0, TCSAFLUSH, &savetty) == -1)
        {
            StatusMsg(MSG_TTY_RST_FAILED);
            exit(16);
        }
}

/*-------------------------------------------------------------------*/
/* Close Socket.                                                     */
/*                                                                   */
/* Call With:  int     - a pointer to a socket structure.            */
/*                                                                   */
/* Returns:    void                                                  */
/*-------------------------------------------------------------------*/

void CloseSocks(int sock)
{
    if (close(sock))                // Close UDP socket.
    {
        SockErr(ERR_SOCKET_CLOSE);  // Close socket failed message.
        exit(16);                   // Return with RC=16.
    }
    StatusMsg(MSG_SOCKET_CLOSED);   // Socket successfully closed message.
}

/*-------------------------------------------------------------------*/
/* Display Error messages with status code.                          */
/*                                                                   */
/* Call With:  int     - a subscript into the error message table.   */
/*                                                                   */
/* Returns:    void                                                  */
/*-------------------------------------------------------------------*/

void SockErr(int x)
{
    if (!DaemonFlag)
        cout << "Error. "
             << SockErrors[x]
             << " returned error code "
             << error
             << ".\n"
             << strerror(error)
             << endl;
             
    openlog("SNMPLog", LOG_PID, LOG_DAEMON);
    
    syslog(LOG_DAEMON | LOG_ERR, "Error. %s returned error code %0d.",
        SockErrors[x], error);
    syslog(LOG_DAEMON | LOG_ERR, "%s.",
        strerror(error));
    
    closelog();
}

/*-------------------------------------------------------------------*/
/* Open log file.                                                    */
/*                                                                   */
/* Call With:  void                                                  */
/*                                                                   */
/* Returns:    void                                                  */
/*-------------------------------------------------------------------*/

void OpenMsgFile(void)
{
    time_t      tm;                 // Long integer buffer for time.
    char        ts[48];             // Buffer to hold formatted time.
    
    MsgFile = fopen(MSGPATH, "a");

    if (!MsgFile)
    {
        ErrorMsg(MSG_LOG_OPEN_FAIL, errno);
        exit(16);
    }

    time(&tm);
                   
    strftime(ts, sizeof(ts), "%x %X", localtime(&tm));
    
    fprintf(MsgFile, "%s %s %s %s %s\n\n", h1, h2, h3, ts, h1);
    fflush(MsgFile);
}

/*-------------------------------------------------------------------*/
/* Close log file.                                                   */
/*                                                                   */
/* Call With:  void                                                  */
/*                                                                   */
/* Returns:    void                                                  */
/*-------------------------------------------------------------------*/

void CloseMsgFile(void)
{
    time_t      tm;                 // Long integer buffer for time.
    char        ts[48];             // Buffer to hold formatted time.
    
    if (MsgFile)
    {
        time(&tm);
                   
        strftime(ts, sizeof(ts), "%x %X", localtime(&tm));
    
        fprintf(MsgFile, "\n%d %s\n\n%s %s %s %s %s\n",
            iCtr, t1, h1, h2, h4, ts, h1);
    
        if (fclose(MsgFile))
            ErrorMsg(MSG_LOG_CLOSE_FAIL, errno);
    }
}

/*-------------------------------------------------------------------*/
/* Display Status messages.                                          */
/*                                                                   */
/* Call With:  int     - a subscript into the message table.         */
/*                                                                   */
/* Returns:    void                                                  */
/*-------------------------------------------------------------------*/

void StatusMsg(int x)
{
    if (!DaemonFlag)
        cout << Messages[x]
             << endl;

    openlog("SNMPLog", LOG_PID, LOG_DAEMON);
    
    syslog(LOG_DAEMON | LOG_INFO, "%s", Messages[x]);
    
    closelog();
}

/*-------------------------------------------------------------------*/
/* Send error messages to syslog.                                    */
/*                                                                   */
/* Call With:  int     - a subscript into the message table.         */
/*             int     - a valid errno value.                        */
/*                                                                   */
/* Returns:    void                                                  */
/*-------------------------------------------------------------------*/

void ErrorMsg(int x, int e)
{
    openlog("SNMPLog", LOG_PID, LOG_DAEMON);
        
    syslog(LOG_DAEMON | LOG_ERR, "%s\n%s%d.\n%s\n",
           Messages[x],
           Messages[MSG_ERROR_CODE],
           strerror(e));

    closelog();
}

/*-------------------------------------------------------------------*/
/* Parse Command Line Arguments.                                     */
/*                                                                   */
/* Call With:  int     - an integer with the number of command line  */
/*                       arguments (argc).                           */
/*             char ** - a pointer to an array of string arguments.  */
/*                                                                   */
/* Returns:    void    - nothing.                                    */
/*-------------------------------------------------------------------*/
void    ParseArgs(int n, char **p)
{
    int    x      = 1;

    DaemonFlag    = false;
    SyslogFlag    = true;
        
    while (x < n)
    {
        if ((p[x][0] != '-') || (strlen(p[x]) != 2))
        {
            StatusMsg(MSG_INVALID_ARG);
            exit(8);
        }
        
        switch    (p[x][1])
        {
            case    'd':
            case    'D':
                DaemonFlag = true;
                x++;
                break;
            case    's':
            case    'S':
                SyslogFlag = false;
                x++;
                break;
            case    'f':
            case    'F':
            {
                    char    c;

                    iCMDfd = open(CMDPATH, O_WRONLY | O_NONBLOCK);

                    if (iCMDfd == -1)
                    {
                        cout << Messages[MSG_PIPE_OP_FAIL]
                             << '\n'
                             << Messages[MSG_ERROR_CODE]
                             << errno
                             << ".\n"
                             << strerror(errno)
                             << '.'
                             << endl;
                        exit(16);
                    }

                    c = 'f';

                    if (write(iCMDfd, &c, 1) == -1)
                    {
                        cout << Messages[MSG_PID_OPEN_FAIL]
                             << '\n'
                             << Messages[MSG_ERROR_CODE]
                             << errno
                             << ".\n"
                             << strerror(errno)
                             << '.'
                             << endl;
                    }

                    close(iCMDfd);
                    exit(0);
            }
            case    'q':
            case    'Q':
                QuitProcess();
                exit(0);
            case    'p':
            case    'P':
                cout << Messages[MSG_CURRENT_PID]
                     << ReadPID()
                     << '.'
                     << endl;
        
                exit(0);
            case    '?':
            case    'h':
            case    'H':
                cout << HelpMessage
                     << endl;
                exit(0);
            case    'v':
            case    'V':
                cout << Version
                     << endl;
                exit(0);
            default:
                StatusMsg(MSG_INVALID_ARG);
                exit(8);
        }
    }
}

/*-------------------------------------------------------------------*/
/* Interpret TTY Input.                                              */
/*                                                                   */
/* Call With:  char    - The character received from the TTY.        */
/*                                                                   */
/* Returns:    bool    - FALSE (0)  Processed character.             */
/*                       TRUE  (!0) Did nothing.                     */
/*-------------------------------------------------------------------*/

bool TTYInput(char c)
{
    switch (c)
    {
        case    'q':
        case    'Q':
            return true;
        case    'p':
        case    'P':
            cout << Messages[MSG_CURRENT_PID]
                 << ReadPID()
                 << '.'
                 << endl;
            break;
        case    '?':
        case    'h':
        case    'H':
            cout << TTYChars
                 << endl;
            break;
        case    'v':
        case    'V':
            cout << Version
                 << endl;
        default:
            break;
    }
    
    return false;
}

/*-------------------------------------------------------------------*/
/* Terminate a running SNMPLog process.                              */
/*                                                                   */
/* Call With:  void    - nothing.                                    */
/*                                                                   */
/* Returns:    void    - nothing.                                    */
/*-------------------------------------------------------------------*/

void QuitProcess(void)
{
    if (kill(ReadPID(), SIGTERM) == -1)
    {
        cout << Messages[MSG_SIGNAL_FAIL]
             << '\n'
             << Messages[MSG_ERROR_CODE]
             << errno
             << ".\n"
             << strerror(errno)
             << '.'
             << endl;
                
        exit(16);
    }
    
    cout << Messages[MSG_DAEMON_STOP]
         << endl;
                
    exit(0);
}

/*-------------------------------------------------------------------*/
/* Install UNIX signal handlers.                                     */
/*                                                                   */
/* Call With:  int     - The signal to handle.                       */
/*                                                                   */
/* Returns:    void    - Nothing.                                    */
/*-------------------------------------------------------------------*/

void SigHandler(int s)
{
    CloseSocks(iUDPSock);           // Close UDP socket.
    StatusMsg(MSG_STOP);            // Display status message.
    
    if (DaemonFlag)
    {
        if (SyslogFlag)
        {
            openlog("SNMPLog", LOG_PID, LOG_DAEMON);
    
            syslog(LOG_DAEMON | LOG_NOTICE,
                "%s", Messages[MSG_DAEMON_STOP]);

            closelog();
        }
    
        if (iCMDfd > 0)
            close(iCMDfd);

        CloseMsgFile();
        unlink(CMDPATH);
    }
    else
        if (tcsetattr(0, TCSAFLUSH, &savetty) == -1)
        {
            StatusMsg(MSG_TTY_RST_FAILED);
            exit(16);
        }

    UnlinkPID();                    // Remove the PID file for this process.
    exit(0);
}

/*-------------------------------------------------------------------*/
/* Create a named pipe for reading run-time commands.                */
/*                                                                   */
/* Call With:  void    - Nothing.                                    */
/*                                                                   */
/* Returns:    void    - Nothing.                                    */
/*-------------------------------------------------------------------*/

void CreatePipe(void)
{
    if (mknod(CMDPATH, S_IFIFO | 0660, 0))
    {
        if (errno == EEXIST)
            return;

        ErrorMsg(MSG_PIPE_CR_FAIL, errno);

        SigHandler(0);
    }

    StatusMsg(MSG_PIPE_CREATED);
}
