#define VERSION  "1.2.5 (Beta)"

#ifndef PIDDIR
#define PIDDIR   "/var/run/"
#endif

#ifndef PIDFILE
#define PIDFILE  "snmplog.pid"
#endif

#define PIDPATH  PIDDIR PIDFILE

#ifndef MSGDIR
#define MSGDIR   "/var/log/"
#endif

#ifndef MSGFILE
#define MSGFILE  "snmplog.messages"
#endif

#define MSGPATH  MSGDIR MSGFILE

#ifndef CMDDIR
#define CMDDIR   "/var/run/"
#endif

#ifndef CMDFILE
#define CMDFILE  "snmplog.cmdpipe"
#endif

#define CMDPATH  CMDDIR CMDFILE

static
char    Copyright[] =
{
    "SNMP (Simple Network Monitoring Protocol) Receiver.\n\n"
    "Copyright  2000-2002, Marc Niegowski\n"
    "Systems Programmer At Large\n"
    "Connectivity, Inc.\n"
    "23 W. Fourth Street\n"
    "Media, PA  19063-2805 USA\n\n"
    "Phone: (610) 566-0227\n"
    "Fax:   (610) 566-0641\n\n"
    "Email: Marc@Tech-Center.com\n"
    "HTTP:  http://www.marcsweb.com\n"
    "FTP:   ftp://ftp.marcsweb.com\n"
    "TELNET:marcsweb.com\n"
};

static
char    Version[] =
{
    "\nSNMPLog Version "
    VERSION
    "\nLinksys EtherFast Cable/DSL Router Log Daemon\n"
};

static
char    TTYChars[] =
{
    "\nSNMPLog TTY Console operations:\n\n"
    "V - Display SNMPLog version.\n"
    "H - Display this brief help text.\n"
    "? - Same as \"H\".\n"
    "P - Display the Process ID (PID) of this SNMPLog process.\n"
    "Q - Terminate SNMPLog.\n"
};

static
char    HelpMessage[] =
{
    "\nSNMPLog Linksys EtherFast Cable/DSL Router Log Daemon\n\n"
    "Usage: snmplogd [option]\n\n"
    "With no options, snmplogd runs in console mode.\n\n"
    "Options:\n"
    "-d    Start snmplogd as a daemon.\n"
    "-D    Same as -d.\n"
    "-f    Flush the SNMPLog message buffer.\n"
    "-F    Same as -f.\n"
    "-p    Display the Process ID (PID) of the current snmplogd daemon.\n"
    "-P    Same as -p.\n"
    "-q    Quit/Terminate the currently running snmplogd daemon.\n"
    "-Q    Same as -q.\n"
    "-s    Suppress sending log flushed messages to syslog.\n"
    "      Can only be used with -d, otherwise ignored.\n"
    "-S    Same as -s.\n"
    "-v    Display snmplogd version information.\n"
    "-V    Same as -v.\n"
    "-h    Display this help text.\n"
    "-H    Same as -h.\n"
    "-?    Same as -h.\n"
};

static
char    RunMessage[] =
{
    "\nSNMPLog PID file \""PIDPATH"\" already exists. Did snmplog die?\n"
};

void    StartDaemon(void);                          // Start Daemon if requested.
void    SavePID(int);                               // Save the PID of the process.
void    GetHost(char *, hostent **);                // Get UDP receiver host.
void    OpenSocks(int *);                           // Open UDP Socket.
void    BindSocks(int *, sockaddr_in *, hostent *); // Bind UDP Socket.
void    Receiver(int);                              // Receiver Socket for UDP.
void    CloseSocks(int);                            // Close UDP Socket.
void    SockErr(int);                               // Display Socket Error Messages.
void    StatusMsg(int);                             // Display Status Messages.
void    ParseArgs(int, char **);                    // Parse Command Line Args.
void    QuitProcess(void);                          // Terminate a running SNMPLog.
bool    TTYInput(char);                             // Interpret TTY Input.
int     ReadPID(void);                              // Return saved SNMPLog daemon PID.
void    UnlinkPID(void);                            // Remove PID file for this process.
void    SigHandler(int);                            // UNIX Signal handler.
void    OpenMsgFile(void);                          // Open log file.
void    CloseMsgFile(void);                         // Close log file.
void    CreatePipe(void);                           // Create a command pipe.
void    ErrorMsg(int, int);                         // Send an error message to syslog.

int     error;                                      // Global error value.
bool    DaemonFlag;                                 // Global daemon flag.
bool    SyslogFlag;                                 // Global syslog flag.
int     PIDfd;                                      // PID file descriptor.

FILE   *MsgFile;                                    // Log file stream.
termios savetty;                                    // TTY attributes.  

#define MSG_START           0
#define MSG_STOP            1
#define MSG_SOCKET_OPENED   2
#define MSG_SOCKET_BOUND    3
#define MSG_SOCKET_CLOSED   4
#define MSG_RCVR_FOUND      5
#define MSG_TTY_ADDR_FAILED 6
#define MSG_TTY_RAW_FAILED  7
#define MSG_TTY_RST_FAILED  8
#define MSG_INVALID_ARG     9
#define MSG_DAEMON_START    10
#define MSG_DAEMON_STOP     11
#define MSG_DAEMON_FAIL     12
#define MSG_ERROR_CODE      13
#define MSG_PID_CREATE_FAIL 14
#define MSG_PID_UNLINK_FAIL 15
#define MSG_PID_OPEN_FAIL   16
#define MSG_CURRENT_PID     17
#define MSG_SIGNAL_FAIL     18
#define MSG_LOG_OPEN_FAIL   19
#define MSG_LOG_FLUSHED     20
#define MSG_LOG_FLUSH_FAIL  21
#define MSG_PIPE_CR_FAIL    22
#define MSG_LOG_CLOSE_FAIL  23
#define MSG_PIPE_OP_FAIL    24
#define MSG_PIPE_RD_FAIL    25
#define MSG_PIPE_WR_FAIL    26
#define MSG_PIPE_CREATED    27
#define MSG_PIPE_SL_FAIL    28
    
char *Messages[] =                                  // Message Table.
{
    "SNMPLog Started.",                             // Message 0.
    "SNMPLog ended.",                               // Message 1.
    "UDP Socket successfully opened.",              // Message 2.
    "UDP Socket successfully bound.",               // Message 3.
    "UDP Socket successfully closed.",              // Message 4.    
    "SNMP receiver found.",                         // Message 5.
    "Error. TTY tcgetaddr function call failed.",   // Message 6.
    "Error. Unable to put TTY into RAW mode.",      // Message 7.
    "Error. Unable to restore TTY attributes.",     // Message 8.
    "Error. Invalid argument.  Try -h or nothing.", // Message 9.
    "SNMPLog UDP Receiver daemon started.",         // Message 10.
    "SNMPLog UDP Receiver daemon ended.",           // Message 11.
    "SNMPLog UDP Receiver daemon failed to start.", // Message 12.
    "Error code: ",                                 // Message 13.
    "Unable to create the SNMPLog PID file \""      // Message 14.
    PIDPATH "\".",                                    
    "Unable to remove SNMPLog PID file \""          // Message 15.
    PIDPATH "\".",
    "Unable to open SNMPLog PID file \""            // Message 16.
    PIDPATH "\".",
    "Currently running SNMPLog PID = ",             // Message 17. 
    "SNMPLog UDP Receiver daemon kill failed.",     // Message 18.
    "Unable to open SNMPLog log file \""            // Message 19.
    MSGPATH "\".",
    "SNMPLog buffer flushed to log file \""         // Message 20.
    MSGPATH "\".",
    "Unable to flush SNMPLog log file.",            // Message 21.
    "Unable to create command pipe \""              // Message 22.
    CMDPATH "\".",
    "Log file \"" MSGPATH "\" close failure.",      // Message 23.
    "Unable to open command pipe \""                // Message 24.
    CMDPATH "\".",
    "Unable to read command pipe \""                // Message 25.
    CMDPATH "\".",
    "Unable to write to command pipe \""            // Message 26.
    CMDPATH "\".",
    "SNMPLog command pipe created.",                // Message 27.
    "Unable to select pipe \"" CMDPATH "\"."        // Message 28.
};

#define ERR_SOCKET_OPEN     0
#define ERR_SOCKET_BIND     1
#define ERR_SOCKET_SELECT   2
#define ERR_SOCKET_RECEIVE  3
#define ERR_SOCKET_CLOSE    4

char *SockErrors[] =                                // Socket Error Message Table.
{
    "UDP socket",                                   // Socket Error Message 0.
    "UDP Socket bind",                              // Socket Error Message 1.
    "UDP Socket select",                            // Socket Error Message 2.
    "UDP Socket recvfrom",                          // Socket Error Message 3.
    "Socket close"                                  // Socket Error Message 4.
};

#define BUFFER_SIZE         512                     // Receiver UDP Buffer Size.
#define RAW_DATA            0x0049                  // Offset into datagram.

int    SIGNAL_HANDLERS[] =
{
    SIGHUP,
    SIGINT,
    SIGQUIT,
    SIGABRT,
    SIGPIPE,
    SIGALRM,
    SIGTERM,
    SIGUSR1,
    SIGUSR2,
    SIGXCPU,
    SIGXFSZ,
    0
};

int    iUDPSock;
int    iCtr;
int    iCMDfd;

static
char   *h1  = "----------";
static
char   *h2  = "Log File";
static
char   *h3  = "Opened";
static
char   *h4  = "Closed";
static
char   *t1  = "Log Records Written.";

