/*-------------------------------------------------------------------*/
/* xsnmplog                                                          */
/* ========                                                          */
/*                                                                   */
/* X log viewer and interface for the snmplogd daemon.               */
/*                                                                   */
/* Copyright  2002                                                  */
/* Marc Niegowski                                                    */
/* Systems Programmer at large                                       */
/* Connectivity, Inc.                                                */
/* 23 W. Fourth Street                                               */
/* Media, PA 19063-2805                                              */
/* USA                                                               */
/*                                                                   */
/* Phone: 610-566-0227                                               */
/* Fax:   610-566-0641                                               */
/*                                                                   */
/* Email: Marc@Tech-Center.com                                       */
/* Web:   http://www.marcsweb.com/                                   */
/* Ftp:   ftp://ftp.marcsweb.com/                                    */
/*-------------------------------------------------------------------*/

#include <gtk/gtk.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <xsnmplog.h>

int main(int argc, char **argv, char **envp)
{
    GtkWidget *win;                // Window widget.
    GtkWidget *icon;               // Icon widget.
    GtkWidget *vbox;               // Vertical packing box widgets.
    GtkWidget *hbox;               // Horizontal packing box widgets.
    GtkWidget *bbox;               // Buttonbox Widget.
    GtkWidget *frame;              // Frame Widgets.
    GtkWidget *clist;              // Column list wdiget.
    GtkWidget *button_exit,        // Exit button.
              *button_flush;       // Flush button.
    GtkWidget *scroll;             // Scrolling widget.
    GtkWidget *label;              // Label widgets.

    GtkStyle  *style;
    GdkPixmap *pixmap;
    GdkBitmap *bitmap;

    int        x;

    char      *heading   =
               "xsnmplog version " VERSION;

    char      *titles[2] =
               {
                "I/O Date & Time",
                "SNMP Trap Message"
               };

    data     gpdata; 

    gpdata.label_start = "Start";
    gpdata.label_stop  = "Stop";

    gtk_init(&argc, &argv);        // Initialize gtk+.

    colormap         = gdk_colormap_get_system();

    for (x = 0; x < 4; x++)
    {
        color[x].red   = 0;
        color[x].blue  = 0;
        color[x].green = 0;
    }
    
    color[COLOR_R].red   = 0xffff;
    color[COLOR_G].green = 0x7fff;
    color[COLOR_B].blue  = 0xffff;
/*
    color[COLOR_K].red   = 0xffff;
    color[COLOR_K].green = 0xffff;
    color[COLOR_K].blue  = 0xffff;
*/
    for (x = 0; x < 4; x++)
        if (!gdk_color_alloc(colormap, &color[x]))
            printf("Unable to allocate color.\n");

    win = gtk_window_new(GTK_WINDOW_TOPLEVEL);

    gtk_signal_connect(GTK_OBJECT(win), "delete_event",
                       GTK_SIGNAL_FUNC(event_delete), NULL);

    gtk_signal_connect(GTK_OBJECT(win), "destroy",
                       GTK_SIGNAL_FUNC(event_destroy), NULL);

    gtk_window_set_title(GTK_WINDOW(win), heading);

    gtk_widget_set_usize(GTK_WIDGET(win), WINDOW_WIDTH, WINDOW_HEIGHT);

    gtk_widget_show(win);

    style = gtk_widget_get_style(win);

    vbox  = gtk_vbox_new(FALSE, 5);

    gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
    gtk_container_add(GTK_CONTAINER(win), vbox);

    frame = gtk_frame_new("Messages");
    label = gtk_label_new(NULL);

    gtk_misc_set_alignment(GTK_MISC(label), 0, 0);
    gtk_misc_set_padding(GTK_MISC(label), 5, 1);

    gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 5);
    gtk_container_add(GTK_CONTAINER(frame), label);

    scroll= gtk_scrolled_window_new(NULL, NULL);
    clist = gtk_clist_new_with_titles(2, titles);

    gtk_clist_set_shadow_type(GTK_CLIST(clist), GTK_SHADOW_OUT);
    gtk_clist_set_column_width(GTK_CLIST(clist), 0, COLUMN_WIDTH);

    gtk_box_pack_start(GTK_BOX(vbox), scroll, TRUE, TRUE, 0);
    gtk_container_add(GTK_CONTAINER(scroll), clist);

    hbox  = gtk_hbox_new(FALSE, 5);
    frame = gtk_frame_new("Daemon Control");
    bbox  = gtk_hbutton_box_new();

    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

    gtk_container_set_border_width(GTK_CONTAINER(bbox), 5);

    gtk_container_add(GTK_CONTAINER(hbox), frame);
    gtk_container_add(GTK_CONTAINER(frame), bbox);

    button_exit       = gtk_button_new_with_label("Exit");
    button_flush      = gtk_button_new_with_label("Flush");
    gpdata.button_run = gtk_button_new_with_label("");
    gpdata.checkbutton= gtk_check_button_new_with_label("Auto\nScroll");
    gpdata.autoscroll = 1;
    gpdata.scroll     = 0;

    gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(gpdata.checkbutton),
                               TRUE);

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gpdata.checkbutton),
                                 gpdata.autoscroll);

    pixmap = gdk_pixmap_create_from_xpm(win->window,
                                        &bitmap,
                                        &style->bg[GTK_STATE_NORMAL],
                                        ICNPATH);

    gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox),
                              GTK_BUTTONBOX_START);

    gtk_container_add(GTK_CONTAINER(bbox), button_exit);
    gtk_container_add(GTK_CONTAINER(bbox), button_flush);
    gtk_container_add(GTK_CONTAINER(bbox), gpdata.button_run);
    gtk_container_add(GTK_CONTAINER(bbox), gpdata.checkbutton);

    if (pixmap)
    {
        icon   = gtk_pixmap_new(pixmap, bitmap);

        gtk_container_add(GTK_CONTAINER(hbox), icon);
    }
    
    gtk_signal_connect(GTK_OBJECT(button_exit),       "clicked",
                       GTK_SIGNAL_FUNC(event_destroy),
                       NULL);

    gtk_signal_connect(GTK_OBJECT(button_flush),      "clicked",
                       GTK_SIGNAL_FUNC(event_flush),
                      (gpointer) &gpdata);

    gtk_signal_connect(GTK_OBJECT(gpdata.button_run), "clicked",
                       GTK_SIGNAL_FUNC(event_run),
                      (gpointer) &gpdata);

    gtk_signal_connect(GTK_OBJECT(gpdata.checkbutton),"toggled",
                       GTK_SIGNAL_FUNC(event_autoscroll),
                      (gpointer) &gpdata);


    memset(gpdata.buffer, 0, sizeof(gpdata.buffer));

    gpdata.clist        = clist;
    gpdata.bbox         = bbox;

    gpdata.fd           = open(LOGFILE, O_RDONLY | O_NONBLOCK);

    if (gpdata.fd == -1)
    {
        printf("Error: log file open returned error code %d.\n", errno);
        printf("Error: %s.\n", strerror(errno));
        exit(16);
    }

    if (pipe(pipefd) == -1)
    {
        printf("Error: pipe returned error code %d.\n", errno);
        printf("Error: %s.\n", strerror(errno));
        exit(16);
    }

    gpdata.label        = label;
    counter             = 0;

    update(&gpdata);

    gpdata.tag    = gtk_timeout_add(10, timer, &gpdata);

    gtk_widget_show_all(win);

    gtk_main();

    close(gpdata.fd);
    close(pipefd[0]);
    close(pipefd[1]);

    exit(0);                       // Return to caller.
}

gint timer(gpointer d)
{
    GtkWidget *child; 
    char       output[BUFFER_SIZE],
               number[32],
               message[BUFFER_SIZE],
              *p1,
              *p2;

    data      *gpdata;

    gpdata    = d;

    gtk_timeout_remove(gpdata->tag);

    output[0] = 0;
    p1        = message;

    forkcmd('p', output, sizeof(output));

    if (strstr(output, "Unable "))
    {
        if (gpdata->label)
        {
            strcpy(message, CMDNAME " daemon is not running\n");
            p1 = message;
        }

        if (gpdata->button_run)
        {
            child = GTK_BIN(gpdata->button_run)->child;
            
            if (GTK_IS_LABEL(child))
                gtk_label_set_text(GTK_LABEL(child), gpdata->label_start);
        }
    }
    else
    {
        if (gpdata->label)
        {
            p1 = strchr(output, '=');

            if (p1)
            {
                p1 += 2;
                p2  = strchr(p1, '.');

                if (p2)
                    *p2 = '\0';
                strcpy(message, CMDNAME " daemon is running (PID ");
                strcat(message, p1);
                strcat(message, ")\n");
                p1 = message;
            }
            else
                p1 = output;
        }

        if (gpdata->button_run)
        {
            child = GTK_BIN(gpdata->button_run)->child;
            
            if (GTK_IS_LABEL(child))
                gtk_label_set_text(GTK_LABEL(child), gpdata->label_stop);
        }
    }

    update(gpdata);

    if (gpdata->label)
    {
        sprintf(number, "%d", counter);
        editnum(number, sizeof(number) - 1);
        strcat(p1, number);
        strcat(p1, " messages buffered");
        gtk_label_set_text(GTK_LABEL(gpdata->label), p1);
        gtk_label_set_justify(GTK_LABEL(gpdata->label), GTK_JUSTIFY_LEFT);
    }

    gpdata->tag  = gtk_timeout_add(TIMER * 1000, timer, (gpointer) gpdata);

    return(TRUE);
}

void update(data *gpdata)
{
    char   *element[2],
           *p;
    int     x, l;

//  gtk_clist_freeze(GTK_CLIST(gpdata->clist));

    element[0] = gpdata->buffer;

    x = -1;

    while(x)
    {
        x = readlog(gpdata->fd, (char *) gpdata->buffer);

        if (x < 1)
            break;

        if (gpdata->buffer[0] == '\n') // Discard blank lines.
        {
            gpdata->buffer[0] = 0;
            continue;
        }

        l = strlen(gpdata->buffer);

        if (gpdata->buffer[l - 1] == '\n')
        {
            if (gpdata->buffer[0] == '-')
            {
                gpdata->buffer[0] = 0;
                continue;
            }

            if (strstr(gpdata->buffer, "Log Records"))
            {
                gpdata->buffer[0] = 0;
                continue;
            }

            gpdata->scroll     = 1;

            if (!(p = strchr(gpdata->buffer, ' ')))
            {
                gpdata->buffer[0] = 0;
                continue;
            }

            if ((p + 1) >= (gpdata->buffer + l))
            {
                gpdata->buffer[0] = 0;
                continue;
            }

            if (!(p = strchr(p + 1, ' ')))
            {
                gpdata->buffer[0] = 0;
                continue;
            }

            *p         = 0;
            element[1] = p + 1;

            format(element[1]);
                   
            gtk_clist_append(GTK_CLIST(gpdata->clist), element);

            gtk_clist_set_foreground(GTK_CLIST(gpdata->clist),
                                     counter,
                                     &color[editcolor(element[1])]);
            counter++;

            gpdata->buffer[0]  = 0;
        }
        else
            break;                 // Get out on incomplete records.
    }

//  gtk_clist_thaw(GTK_CLIST(gpdata->clist));

    if ((counter > 0) && (gpdata->scroll))
        if (gpdata->autoscroll)
            gtk_clist_moveto(GTK_CLIST(gpdata->clist),
                             counter - 1, 0, 1, 1);

    gpdata->scroll = 0;
}

int readlog(int fd, char *buf)
{
    int     n, i;
    i = strlen(buf);

    while(i < (BUFFER_SIZE - 1))
    {
        n = read(fd, buf + i, 1);

        if (n == -1)
        {
            printf("Error: read returned error code %d.\n", errno);
            printf("Error: %s.\n", strerror(errno));
            return(n);
        }

        if (!n)
            break;
        else
        {
            i++;

            buf[i] = 0;

            if (buf[i - 1] == '\n')
                break;

            // Truncate any lines that might exceed the buffer.

            if (i == BUFFER_SIZE - 1)
                i--;
        }
    }

    if (i == BUFFER_SIZE - 1)
        buf[i - 1] = '\n';

    return(i);  
}

void format(char *buf)
{
    char   fmdata[BUFFER_SIZE],
          *pointer1,
          *pointer2;

    if (strlen(buf) + 9 > sizeof(fmdata) -1)
       return;

    if (strncmp(buf, "in  ", 3) &&
        strncmp(buf, "out ", 4))
        return;

    if (!strncmp(buf, "in ", 3))
        strcpy(fmdata, "in  ");

    if (!strncmp(buf, "out ", 4))
        strcpy(fmdata, "out ");

    if (!(pointer1 = strchr(buf, ' ')))
        return;

    if (!(pointer2 = strchr(pointer1 + 1, ' ')))
        return;

    strncat(fmdata, pointer1 + 1, pointer2 - pointer1);
    strcat(fmdata, "[");

    pointer1 = pointer2 + 1;

    if (!(pointer2 = strchr(pointer1, ' ')))
        return;

    strncat(fmdata, pointer1, pointer2 - pointer1);
    strcat(fmdata, "] ==> ");

    pointer1 = pointer2 + 1;

    if (!(pointer2 = strchr(pointer1, ' ')))
        return;

    strncat(fmdata, pointer1, pointer2 - pointer1);
    strcat(fmdata, "[");

    pointer1 = pointer2 + 1;

    if (!(pointer2 = strchr(pointer1, 0)))
        return;

    strncat(fmdata, pointer1, pointer2 - pointer1);
    strcat(fmdata, "]");

    strcpy(buf, fmdata);
}

int  editcolor(char *buf)
{
    int     x;

    if (!(strncmp(buf, "in  ", 4)))
        return(COLOR_R);

    if (!(strncmp(buf, "out ", 4)))
        return(COLOR_G);

    x = 0;

    while (gtb1[x])
    {
        if (!(strncmp(buf, gtb1[x], strlen(gtb1[x]))))
            return(COLOR_K);

        x++;
    }

    x = 0;

    while (gtb2[x])
    {
        if (strstr(buf, gtb2[x]))
            return(COLOR_K);

        x++;
    }

    return(COLOR_B);
}

void editnum(char *buf, int max)
{
    int   len,
          rem;

    char  num[32],
         *ptr;

    len    = strlen(buf);

    if (((len / 3) + len) >= sizeof(num))
        return;

    rem    = len % 3;

    rem    = (rem ? rem : 3);

    num[0] = 0;
    ptr    = buf;

    while(len > 0)
    {
        if (len <= 3)
        {
            strcat(num, ptr);
            break;
        }
        else
        {
            strncat(num, ptr, rem);
            strcat(num, ",");
        }

        ptr += rem;
        len -= rem;
        rem  = 3;
    }

    if (strlen(num) > max)
        return;
    else
        strcpy(buf, num);
}
int forkcmd(char a, char *buf, int len)
{
    int      p;
    
    char    *c[3];

    c[0] = CMDPATH;
    c[2] = 0;

    switch(a)
    {
        case   'p':
                c[1] = "-p";
                break;
        case   'f':
                c[1] = "-f";
                break;
        case   'q':
                c[1] = "-q";
                break;
        case   'd':
                c[1] = "-d";
                break;
        default:
                return(0);
    }
      
    p = fork();

    if (p == -1)
    {
        printf("Error: fork returned error code %d.\n", errno);
        printf("Error: %s.\n", strerror(errno));
        return(errno);
    }

    if (p == 0)
    {
        if (buf)
        {
            close(0);              // Close STDIN.   
            close(1);              // Close STDOUT.
            dup(pipefd[0]);        // Dup STDIN pipe.
            dup(pipefd[1]);        // Dup STDOUT pipe.
            close(pipefd[0]);      // Close dup STDIN.
            close(pipefd[1]);      // Close dup STDOUT.
        }

        if (execvp(c[0], c)  == -1)
        {
            printf("Error: execvp returned error code %d.\n", errno);
            printf("Error: %s.\n", strerror(errno));
        }
        _exit(0);
    }
    else
    {        
        struct
        timeval t;

        fd_set  s;

        int     n;

        if (waitpid(p, NULL, 0) == -1)
        {
            printf("Error: waitpid returned error code %d.\n", errno);
            printf("Error: %s.\n", strerror(errno));
            return(errno);
        }

        if (!buf)
            return(0);

        n = -1;

        while(n)
        {
            FD_ZERO(&s);
            FD_SET(pipefd[0], &s);
            t.tv_sec  = 1;
            t.tv_usec = 0;

            n = select(pipefd[0]+1, &s, NULL, NULL, &t);

            if (n == -1)
            {
                printf("Error: select returned error code %d.\n", errno);
                printf("Error: %s.\n", strerror(errno));
                break;
            }

            if (!n)
                break;

            n = read(pipefd[0], buf, len-1);

            if (n == -1)
            {
                printf("Error: read returned error code %d.\n", errno);
                printf("Error: %s.\n", strerror(errno));
                break;
            }

            buf[n] = 0;
        }
    }
   return(0);
}

void event_run(GtkButton *w, gpointer d)
{
    GtkWidget *child;
    data      *gpdata;

    char       output[BUFFER_SIZE];

    gpdata = (data *) d;

    forkcmd('p', output, sizeof(output));

    if (gpdata->button_run)
    {
        if (strstr(output, "Unable "))
        {
            forkcmd('d', output, sizeof(output));
            child = GTK_BIN(gpdata->button_run)->child;
            
            if (GTK_IS_LABEL(child))
                gtk_label_set_text(GTK_LABEL(child), gpdata->label_stop);
        }

        if (strstr(output, "Currently "))
        {
            forkcmd('q', output, sizeof(output));
            child = GTK_BIN(gpdata->button_run)->child;
            
            if (GTK_IS_LABEL(child))
                gtk_label_set_text(GTK_LABEL(child), gpdata->label_start);
        }

    }
    return;
}

void event_flush(GtkButton *w, gpointer d)
{
    forkcmd('f', NULL, 0);

    return;
}

void event_autoscroll(GtkButton *w, gpointer d)
{
    data   *gpdata;

    gpdata = (data *) d;

    gpdata->autoscroll =  (gpdata->autoscroll ? 0 : 1);

    return;
}

int event_delete(GtkWidget *w, GdkEvent *e, gpointer d)
{
    return(FALSE);
}

void event_destroy(GtkWidget *w, gpointer d)
{
    gtk_main_quit();
}
